from typing import Union
from datetime import datetime, timedelta


_FMT = '%Y%m%d%H%M%S.%f'


def dt_from_str(s: str) -> Union[datetime, timedelta]:
    """String to datetime.

    https://www.dmtf.org/sites/default/files/standards/documents/DSP0004V2.3_final.pdf
    """  # nopep8
    s = s.replace('*', '0')

    if s.endswith(':000'):
        # ddddddddhhmmss.mmmmmm:000
        return timedelta(
            days=int(s[:8]),
            hours=int(s[8:10]),
            minutes=int(s[10:12]),
            seconds=int(s[12:14]),
            microseconds=int(s[15:21])
        )

    # timestamp: yyyymmddhhmmss.mmmmmmsutc
    # utc is offset in minutes
    minutes = int(s[-3:])
    hours = minutes // 60
    minutes = minutes % 60

    for t in (4, 6):
        # months and days must start at 1, not 0
        e = t + 2
        if s[t:e] == '00':
            s = s[:t] + '01' + s[e:]

    dt = datetime.strptime(s[:21], _FMT)
    if s[-4] == '+':
        dt -= timedelta(hours=hours, minutes=minutes)
    else:
        assert s[-4] == '-'
        dt += timedelta(hours=hours, minutes=minutes)
    return dt


if __name__ == '__main__':
    print(dt_from_str('20220207094949.500000+060'))
    print(dt_from_str('19980525133015.0000000-300'))
    print(dt_from_str('19980525183015.0000000+000'))
    print(dt_from_str('19980525******.0000000+000'))
    print(dt_from_str('1998**********.0000000+000'))
    print(dt_from_str('00000001132312.125***:000'))



"""
ORPCTHIS

version
\x05\x00\x07\x00
flags
\x01\x00\x00\x00
reserverd
\x00\x00\x00\x00
cid
\xcf=\xd6\x19\xff\xf5\xfb\x11\xc5\xbc\x85[\xec*\xc9$
extensions (4 / 8)
\x00\x00\x00\x00

p_unk_outer
\x00\x00\x00\x00

referent_id (4) random
\xc2h\x00\x00

ulCntData (416)
\xa0\x01\x00\x00

again??? (4)
\xa0\x01\x00\x00

OBJ REF

signature (4)
MEOW

flags (4)
\x04\x00\x00\x00

OBJ REF CUSTOM

iid (16)
\xa2\x01\x00\x00\x00\x00\x00\x00\xc0\x00\x00\x00\x00\x00\x00F

clsid (16)
8\x03\x00\x00\x00\x00\x00\x00\xc0\x00\x00\x00\x00\x00\x00F

extensions (4)
\x00\x00\x00\x00

object_ref_size (4) (376 = below + 8)
x\x01\x00\x00

ActivationBLOB

h\x01\x00\x00\x00...

padding
\xfa\xfa\xfa\xfa\xfa\xfa'
"""
import struct
from .activation_blob import ActivationBlob
from .activation_context_info_data import ActivationContextInfoData
from .instantiation_info_data import InstantiationInfoData
from .location_info_data import LocationInfoData
from .activation_context_info_data import ActivationContextInfoData
from .scm_request_info_data import ScmRequestInfoData
from .orpcthis import ORPCTHIS
from ..tools import gen_referent_id
from .objref_custom import ObjRefCustom


class RemoteCreateInstance:

    FMT32 = '<LLLL'
    FMT64 = '<QQLL'

    def __init__(self, class_id: bytes, iid: bytes):

        activatoin_blob = ActivationBlob()

        activatoin_blob.add_info_data(InstantiationInfoData(class_id, iid))
        activatoin_blob.add_info_data(LocationInfoData)
        activatoin_blob.add_info_data(ActivationContextInfoData)
        activatoin_blob.add_info_data(ScmRequestInfoData)

        obj_ref_custom = ObjRefCustom.init()
        obj_ref_custom.set_object(activatoin_blob.get_data())

        self.ab_data = obj_ref_custom.get_data()

    def get_data(self) -> bytes:

        p_unk_outer = 0
        referent_id = gen_referent_id()
        c_ab_data = len(self.ab_data)

        data = ORPCTHIS.get_data(flags=1) + struct.pack(
            self.FMT32,
            p_unk_outer,
            referent_id,
            c_ab_data,
            c_ab_data,
        ) + self.ab_data

        return data



import struct
from typing import OrderedDict
from .orpcthat import ORPCTHAT
from .objref_standard import ObjRefStandard
from .activation_blob import ActivationBlob
from .scm_reply_info_data import ScmReplyInfoData
from .props_out_info import PropsOutInfo
from ..tools import is_fqdn
from .varying_array import VaryingArray
from .encoding_unit import EncodingUnit
from .property_info import PropertyInfo


class NextResponse:

    _FMT1_32 = '<LLL'
    _FMT1_64 = '<QQQ'

    _FMT1_32_SZ = struct.calcsize(_FMT1_32)
    _FMT1_64_SZ = struct.calcsize(_FMT1_64)

    def __init__(self, data: bytes):
        self.orpcthat, offset = ORPCTHAT.from_data(data, offset=0)
        (
            ncount,
            noffset,
            ncount,
        ) = struct.unpack_from(self._FMT1_32, data, offset)
        offset += self._FMT1_32_SZ
        self._ap_objects = []

        for _ in range(ncount):
            va, offset = VaryingArray.from_data(data, offset)
            encoding_unit = EncodingUnit(va.objref.object_data)
            self._ap_objects.append(encoding_unit)

        (
            self.pu_returned,
            self.error_code,
        ) = struct.unpack_from('<LL', data, offset)

    def get_properties(
            self,
            ignore_defaults: bool = False,
            ignore_missing: bool = False,
            load_qualifiers: bool = False) -> OrderedDict[str, PropertyInfo]:
        """Get properties
        ignore_defaults: Ignore default values. Set missing values to None
                         if a value does not exist in the current class.
                         ignore_defaults will always be True if ignore_missing
                         is set to True.
        ignore_missing: If set to True, values missing in the current class
                        will not be part of the result.
        load_qualifiers: Load the qualifiers of the properties. If not, the
                         property qualifier_set will have the offset in the
                        heap where the qualifiers are stored.
        """

        items = []
        assert len(self._ap_objects) == 1
        for encoding_unit in self._ap_objects:
            object_block = encoding_unit.object_block
            class_part = object_block.class_part
            properties = class_part.properties

            if not ignore_defaults:
                properties.set_prop_defaults(class_part.nd_value_table)
                properties.set_prop_values(
                    class_part.class_heap,
                    class_part.nd_value_table,
                    set_defaults=True)

            properties.set_prop_values(
                object_block.instance_heap,
                object_block.nd_value_table,
                ignore_missing=ignore_missing)

            if load_qualifiers:
                properties.set_qualifiers(class_part.class_heap)

            return properties.properties



"""

\x01\x10\x08\x00\xcc\xcc\xcc\xcc
\x1a\x00\x00\x00\xcc\xcc\xcc\xcc

dw_reserved (4)
\x00\x00\x00\x00

ReferentId (4)
<\x15\x00\x00

ClientImpLevel (4)
\x00\x00\x00\x00

cRequestedProtseqs (2)
\x01\x00

padding (2)
\xaa\xaa

ReferentId (4)
\x1cR\x00\x00

Referent Count (4)
\x01\x00\x00\x00

Referent (2)
\x07\x00'
"""
import struct
import random
from .common import NdrCommon
from ..tools import gen_referent_id
from ..dcom_const import CLSID_ScmRequestInfo


class ScmRequestInfoData(NdrCommon):

    CLSID = CLSID_ScmRequestInfo

    BUFFER32_FMT = '<lLLH'
    BUFFER64_FMT = '<lQLH'

    REFERENT32_FMT = '<LL'
    REFERENT64_FMT = '<LQ'

    @classmethod
    def get_data(cls) -> bytes:
        p_requested_prot_seqs = [7]

        dw_reserved = 0
        referent_id = gen_referent_id()
        client_imp_level = 0
        c_requested_prot_seqs = 1

        referents = b''.join([
            struct.pack('<H', x)
            for x in p_requested_prot_seqs])

        padding = b'\xaa\xaa'  # can be fiex as this is aways the same

        buffer = struct.pack(
            cls.BUFFER32_FMT,
            dw_reserved,            # <l
            referent_id,            # <L (32) / <Q (64)
            client_imp_level,       # <L
            c_requested_prot_seqs,  # <H
        ) + padding

        referent_id = gen_referent_id()

        buffer += struct.pack(
            cls.REFERENT32_FMT,
            referent_id,
            len(p_requested_prot_seqs),
        ) + referents

        private, padding = cls.get_private(buffer)

        return cls.get_common() + private + buffer + padding



"""
ORPCTHIS

version
\x05\x00\x07\x00
flags
\x00\x00\x00\x00
reserverd
\x00\x00\x00\x00
cid
\xab\xbeX5\x81:\x82q\x80\xed\x8dq\x05\x84\xa8U
extensions
\x00\x00\x00\x00

cPublicInterfaces
\x01\x00

padding
\xce\xce

cPublicRefs ( <l )
\x01\x00\x00\x00

ipid
\x05p\x00\x00\xb8\x0b\x00\x00E\x1f\xaa\xa1\x8b\x16:\xc3

cMaxCount ( <L )
\x01\x00\x00\x00

cPrivateRefs ( <l )
\x00\x00\x00\x00


"""

import struct
from .activation_blob import ActivationBlob
from .activation_context_info_data import ActivationContextInfoData
from .instantiation_info_data import InstantiationInfoData
from .location_info_data import LocationInfoData
from .activation_context_info_data import ActivationContextInfoData
from .scm_request_info_data import ScmRequestInfoData
from .orpcthis import ORPCTHIS
from ..tools import gen_referent_id
from .objref_custom import ObjRefCustom


class RemRelease:

    FMT = '<Ll'

    def __init__(self, ipid: bytes):
        self.ipid = ipid

    def get_data(self) -> bytes:
        # 1 (<H c_interfaces) + padding + 1 (<l c_public_refs)
        fixed = '\x01\x00\xce\xce\x01\x00\x00\x00'
        c_max_count = 1
        c_private_refs = 0

        data = ORPCTHIS.get_data(flags=0) + fixed + self.ipid + struct.pack(
            self.FMT,
            c_max_count,
            c_private_refs,
        )

        return data
